﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Linq;

namespace HeuristicLab.Encodings.RealVectorEncoding {
  [Item("Default", "Initializes the covariance matrix and step size variables.")]
  [StorableClass]
  public class CovarianceMatrixInitializer : SingleSuccessorOperator, IRealVectorOperator, ICMAESInitializer {

    public Type CMAType {
      get { return typeof(CovarianceMatrixStrategyParameters); }
    }

    #region Parameter Properties
    public IValueLookupParameter<IntValue> DimensionParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Dimension"]; }
    }

    public ILookupParameter<IntValue> PopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }

    public ILookupParameter<IntValue> MuParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Mu"]; }
    }

    public ILookupParameter<CovarianceMatrixStrategyParameters> StrategyParametersParameter {
      get { return (ILookupParameter<CovarianceMatrixStrategyParameters>)Parameters["StrategyParameters"]; }
    }
    #endregion

    [StorableConstructor]
    protected CovarianceMatrixInitializer(bool deserializing) : base(deserializing) { }
    protected CovarianceMatrixInitializer(CovarianceMatrixInitializer original, Cloner cloner) : base(original, cloner) { }
    public CovarianceMatrixInitializer()
      : base() {
      Parameters.Add(new ValueLookupParameter<IntValue>("Dimension", "The problem dimension (N)."));
      Parameters.Add(new LookupParameter<IntValue>("PopulationSize", "The population size (lambda)."));
      Parameters.Add(new LookupParameter<IntValue>("Mu", "The number of offspring considered for updating of the strategy parameters."));
      Parameters.Add(new LookupParameter<CovarianceMatrixStrategyParameters>("StrategyParameters", "The strategy parameters for real-encoded CMA-ES."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CovarianceMatrixInitializer(this, cloner);
    }

    public override IOperation Apply() {
      var N = DimensionParameter.ActualValue.Value;
      var lambda = PopulationSizeParameter.ActualValue.Value;
      var mu = (lambda / 2.0);
      var weights = Enumerable.Range(1, (int)mu).Select(x => Math.Log(mu + 0.5) - Math.Log(x)).ToArray();
      mu = Math.Floor(mu);
      var sum = weights.Sum();
      weights = weights.Select(x => x / sum).ToArray();
      var mueff = 1 / weights.Sum(x => x * x);

      var cc = (4 + mueff / N) / (N + 4 + 2 * mueff / N);  // time constant for cumulation for C
      var cs = (mueff + 2) / (N + mueff + 5);  // t-const for cumulation for sigma control
      var c1 = 2 / ((N + 1.3) * (N + 1.3) + mueff);    // learning rate for rank-one update of C
      var cmu = 2 * (mueff - 2 + 1 / mueff) / ((N + 2) * (N + 2) + mueff);  // and for rank-mu update

      var pc = new double[N]; // evolution paths for C
      var ps = new double[N]; // evolution paths for sigma
      var B = new double[N, N]; // B defines the coordinate system
      for (int i = 0; i < N; i++) B[i, i] = 1;
      var D = new double[N]; // diagonal D defines the scaling
      for (int i = 0; i < N; i++) D[i] = 1;
      var C = (double[,])B.Clone(); // covariance matrix C
      var invsqrtC = (double[,])C.Clone(); // C^-1/2
      var chiN = Math.Sqrt(N) * (1 - 1 / (4 * N) + 1 / (21 * N * N)); // expectation of  ||N(0,I)|| == norm(randn(N,1))

      MuParameter.ActualValue = new IntValue((int)mu);
      StrategyParametersParameter.ActualValue = new CovarianceMatrixStrategyParameters() {
        B = new DoubleMatrix(B),
        C = new DoubleMatrix(C),
        C1 = new DoubleValue(c1),
        CC = new DoubleValue(cc),
        ChiN = new DoubleValue(chiN),
        CMU = new DoubleValue(cmu),
        CS = new DoubleValue(cs),
        D = new DoubleArray(D),
        InvSqrtC = new DoubleMatrix(invsqrtC),
        Mu = new IntValue((int)mu),
        MuEff = new DoubleValue(mueff),
        PC = new DoubleArray(pc),
        PS = new DoubleArray(ps),
        Sigma = new DoubleValue(0.5),
        Weights = new DoubleArray(weights)
      };
      return base.Apply();
    }
  }
}