﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;
using System;

namespace HeuristicLab.Encodings.RealVectorEncoding {
  [Item("CovarianceMatrixMutator", "Mutates the solution vector according to the CMA-ES scheme.")]
  [StorableClass]
  public class CovarianceMatrixMutator : SingleSuccessorOperator, IStochasticOperator, IRealVectorOperator, ICMAESManipulator {

    public Type CMAType {
      get { return typeof(CovarianceMatrixStrategyParameters); }
    }

    #region Parameter Properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public ILookupParameter<RealVector> RealVectorParameter {
      get { return (ILookupParameter<RealVector>)Parameters["RealVector"]; }
    }

    public ILookupParameter<CovarianceMatrixStrategyParameters> StrategyParametersParameter {
      get { return (ILookupParameter<CovarianceMatrixStrategyParameters>)Parameters["StrategyParameters"]; }
    }
    #endregion

    [StorableConstructor]
    protected CovarianceMatrixMutator(bool deserializing) : base(deserializing) { }
    protected CovarianceMatrixMutator(CovarianceMatrixMutator original, Cloner cloner) : base(original, cloner) { }
    public CovarianceMatrixMutator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new LookupParameter<RealVector>("RealVector", "The solution vector of real values."));
      Parameters.Add(new LookupParameter<CovarianceMatrixStrategyParameters>("StrategyParameters", "The CMA-ES strategy parameters used for mutation."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CovarianceMatrixMutator(this, cloner);
    }

    public override IOperation Apply() {
      var random = RandomParameter.ActualValue;
      var arx = RealVectorParameter.ActualValue;
      var sp = StrategyParametersParameter.ActualValue;

      var nd = new NormalDistributedRandom(random, 0, 1);

      var artmp = new double[arx.Length];
      for (int i = 0; i < arx.Length; ++i)
        artmp[i] = sp.D[i] * nd.NextDouble();

      for (int i = 0; i < arx.Length; i++) {
        var sum = 0.0;
        for (int j = 0; j < arx.Length; j++)
          sum += sp.B[i, j] * artmp[j];
        arx[i] += sp.Sigma.Value * artmp[i]; // m + sig * Normal(0,C) 
      }
      return base.Apply();
    }
  }
}