﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Linq;

namespace HeuristicLab.Encodings.RealVectorEncoding {
  [Item("CovarianceMatrixUpdater", "Updates the covariance matrix and strategy parameters of CMA-ES.")]
  [StorableClass]
  public class CovarianceMatrixUpdater : SingleSuccessorOperator, IRealVectorOperator, ICMAESUpdater, IStochasticOperator, IIterationBasedOperator {

    public Type CMAType {
      get { return typeof(CovarianceMatrixStrategyParameters); }
    }

    #region Parameter Properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public ILookupParameter<CovarianceMatrixStrategyParameters> StrategyParametersParameter {
      get { return (ILookupParameter<CovarianceMatrixStrategyParameters>)Parameters["StrategyParameters"]; }
    }

    public IScopeTreeLookupParameter<RealVector> MeansParameter {
      get { return (IScopeTreeLookupParameter<RealVector>)Parameters["Means"]; }
    }

    public IScopeTreeLookupParameter<RealVector> OffspringParameter {
      get { return (IScopeTreeLookupParameter<RealVector>)Parameters["Offspring"]; }
    }

    public ILookupParameter<IntValue> IterationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Iterations"]; }
    }

    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    #endregion

    [StorableConstructor]
    protected CovarianceMatrixUpdater(bool deserializing) : base(deserializing) { }
    protected CovarianceMatrixUpdater(CovarianceMatrixUpdater original, Cloner cloner) : base(original, cloner) { }
    public CovarianceMatrixUpdater()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new LookupParameter<CovarianceMatrixStrategyParameters>("StrategyParameters", "The strategy parameters of CMA-ES."));
      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("Means", "The old and the new mean.", 2));
      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("Offspring", "The created offspring solutions.", 3));
      Parameters.Add(new LookupParameter<IntValue>("Iterations", "The number of iterations passed."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumIterations", "The maximum number of iterations."));
      MeansParameter.ActualName = "RealVector";
      OffspringParameter.ActualName = "RealVector";
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CovarianceMatrixUpdater(this, cloner);
    }

    public override IOperation Apply() {

      var random = RandomParameter.ActualValue;
      var iterations = IterationsParameter.ActualValue.Value;

      var xold = MeansParameter.ActualValue[0];
      var xmean = MeansParameter.ActualValue[1];
      var offspring = OffspringParameter.ActualValue;

      var sp = StrategyParametersParameter.ActualValue;
      var N = sp.PC.Length;
      var damps = 1 + 2 * Math.Max(0, Math.Sqrt((sp.MuEff.Value - 1.0) / (N + 1)) - 1.0) + sp.CS.Value;

      for (int i = 0; i < sp.PS.Length; i++) {
        var matmul = 0.0;
        for (int j = 0; j < sp.PS.Length; j++) {
          matmul += sp.InvSqrtC[i, j] * (xmean[j] - xold[j]);
        }
        sp.PS[i] = (1 - sp.CS.Value) * sp.PS[i]
                   + Math.Sqrt(sp.CS.Value * (2 - sp.CS.Value) * sp.MuEff.Value) * matmul / sp.Sigma.Value;
      }
      var normPS = sp.PS.Select(x => x * x).Sum();
      var hsig = normPS / Math.Sqrt(1 - Math.Pow(1 - sp.CS.Value, 2 * iterations)) / sp.ChiN.Value < 1.4 + 2.0 / (N + 1) ? 1.0 : 0.0;
      for (int i = 0; i < sp.PC.Length; i++) {
        sp.PC[i] = (1 - sp.CC.Value) * sp.PC[i]
                   + hsig * Math.Sqrt(sp.CC.Value * (2 - sp.CC.Value) * sp.MuEff.Value) * (xmean[i] - xold[i]) / sp.Sigma.Value;
      }
      var c = new double[sp.C.Rows, sp.C.Columns];
      for (int i = 0; i < sp.C.Rows; i++) {
        for (int j = 0; j < sp.C.Columns; j++) {
          c[i, j] = (1 - sp.C1.Value - sp.CMU.Value) * sp.C[i, j]
                       + (1 - hsig) * sp.CC.Value * (2 - sp.CC.Value) * sp.C[i, j];
          for (int k = 0; k < sp.Mu.Value; k++) {
            c[i, j] += sp.CMU.Value * sp.Weights[k] * (offspring[k][i] - xold[i]) * (offspring[k][j] - xold[j]) / (sp.Sigma.Value * sp.Sigma.Value);
          }
          sp.C[i, j] = c[i, j];
        }
      }
      sp.Sigma.Value = sp.Sigma.Value * Math.Exp((sp.CS.Value / damps) * (normPS / sp.ChiN.Value - 1));

      double[] eval;
      double[,] evec;
      if (!alglib.smatrixevd(c, N, 1, false, out eval, out evec))
        throw new InvalidOperationException("Unable to perform eigendecomposition of matrix.");

      sp.B = new DoubleMatrix(evec);
      for (int i = 0; i < sp.D.Length; i++)
        sp.D[i] = Math.Sqrt(eval[i]);

      for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
          sp.InvSqrtC[i, j] = 0;
          var v = sp.B[i, j] * (1 / sp.D[j]);
          for (int k = 0; k < N; k++) {
            sp.InvSqrtC[i, j] += v * sp.B[k, j];
          }
        }
      }
      return base.Apply();
    }
  }
}